from sila2.framework.binary_transfer.binary_transfer_error import BinaryTransferError, BinaryTransferErrorType


class InvalidBinaryTransferUUID(BinaryTransferError):
    def __init__(self, message: str):
        super().__init__(BinaryTransferErrorType.INVALID_BINARY_TRANSFER_UUID, message)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from sila2.framework import FullyQualifiedIdentifier
from sila2.server import FeatureImplementationBase

from .silaservice_types import GetFeatureDefinition_Responses, SetServerName_Responses


class SiLAServiceBase(FeatureImplementationBase, ABC):

    """
    This Feature MUST be implemented by each SiLA Server.

    It specifies Commands and Properties to discover the Features a SiLA Server implements as well as details
    about the SiLA Server, like name, type, description, vendor and UUID.

    Any interaction described in this feature MUST not affect the behaviour of any other Feature.
    """

    @abstractmethod
    def get_ServerName(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
        Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command.

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command.
        """
        pass

    @abstractmethod
    def get_ServerType(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
        It is specified by the implementer of the SiLA Server and MAY not be unique.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
        It is specified by the implementer of the SiLA Server and MAY not be unique.
        """
        pass

    @abstractmethod
    def get_ServerUUID(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
        and remain the same for all times.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
        and remain the same for all times.
        """
        pass

    @abstractmethod
    def get_ServerDescription(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
        Description of the SiLA Server. This should include the use and purpose of this SiLA Server.

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Description of the SiLA Server. This should include the use and purpose of this SiLA Server.
        """
        pass

    @abstractmethod
    def get_ServerVersion(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
        a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
        appended, e.g. “3.19.373_mighty_lab_devices”.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
        a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
        appended, e.g. “3.19.373_mighty_lab_devices”.
        """
        pass

    @abstractmethod
    def get_ServerVendorURL(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
        This URL SHOULD be accessible at all times.
        The URL is a Uniform Resource Locator as defined in RFC 1738.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
        This URL SHOULD be accessible at all times.
        The URL is a Uniform Resource Locator as defined in RFC 1738.
        """
        pass

    @abstractmethod
    def get_ImplementedFeatures(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> List[str]:
        """
          Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
        This list SHOULD remain the same throughout the lifetime of the SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
        This list SHOULD remain the same throughout the lifetime of the SiLA Server.
        """
        pass

    @abstractmethod
    def GetFeatureDefinition(
        self, FeatureIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> GetFeatureDefinition_Responses:
        """
          Get the Feature Definition of an implemented Feature by its fully qualified Feature Identifier.
        This command has no preconditions and no further dependencies and can be called at any time.


          :param FeatureIdentifier: The fully qualified Feature identifier for which the Feature definition shall be retrieved.

          :param metadata: The SiLA Client Metadata attached to the call

          :return:

              - FeatureDefinition: The Feature definition in XML format (according to the Feature Definition Schema).


        """
        pass

    @abstractmethod
    def SetServerName(
        self, ServerName: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> SetServerName_Responses:
        """
          Sets a human readable name to the Server Name Property.Command has no preconditions and
        no further dependencies and can be called at any time.


          :param ServerName: The human readable name to assign to the SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call

        """
        pass



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class InvalidCommandExecutionUUID(FrameworkError):
    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.INVALID_COMMAND_EXECUTION_UUID, message)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

from sila2.features.silaservice import SiLAServiceBase, UnimplementedFeature
from sila2.framework import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer


class SiLAServiceImpl(SiLAServiceBase):
    parent_server: SilaServer

    def __init__(self, parent_server: SilaServer):
        self.parent_server = parent_server

    def GetFeatureDefinition(self, FeatureIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        if FeatureIdentifier not in (f.fully_qualified_identifier for f in self.parent_server.features.values()):
            raise UnimplementedFeature(f"Feature {FeatureIdentifier} is not implemented by this server")

        return self.parent_server.features[FeatureIdentifier.split("/")[-2]]._feature_definition

    def SetServerName(self, ServerName: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> None:
        self.parent_server.server_name = ServerName

    def get_ImplementedFeatures(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> List[str]:
        return [f.fully_qualified_identifier for f in self.parent_server.features.values()]

    def get_ServerName(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return self.parent_server.server_name

    def get_ServerType(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return self.parent_server.server_type

    def get_ServerUUID(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return str(self.parent_server.server_uuid)

    def get_ServerDescription(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return self.parent_server.server_description

    def get_ServerVersion(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return self.parent_server.server_version

    def get_ServerVendorURL(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        return self.parent_server.server_vendor_url



from typing import TYPE_CHECKING

from .authorizationservice_base import AuthorizationServiceBase
from .authorizationservice_errors import InvalidAccessToken
from .authorizationservice_feature import AuthorizationServiceFeature

__all__ = [
    "AuthorizationServiceBase",
    "AuthorizationServiceFeature",
    "InvalidAccessToken",
]

if TYPE_CHECKING:
    from .authorizationservice_client import AuthorizationServiceClient  # noqa: F401

    __all__.append("AuthorizationServiceClient")



from jinja2 import Environment

from sila2.code_generator.template_loader import TemplateLoader


class TemplateEnvironment(Environment):
    def __init__(self):
        super().__init__(loader=TemplateLoader(), autoescape=False)
        self.filters["repr"] = repr
        self.filters["strip"] = str.strip
        self.filters["lower"] = str.lower



from __future__ import annotations

from typing import Optional

from sila2.framework.errors.defined_execution_error import DefinedExecutionError

from .lockcontroller_feature import LockControllerFeature


class InvalidLockIdentifier(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The sent lock identifier is not valid."
        super().__init__(LockControllerFeature.defined_execution_errors["InvalidLockIdentifier"], message=message)


class ServerAlreadyLocked(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The SiLA Server can not be locked because it is already locked."
        super().__init__(LockControllerFeature.defined_execution_errors["ServerAlreadyLocked"], message=message)


class ServerNotLocked(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The SiLA Server can not be unlocked because it is not locked."
        super().__init__(LockControllerFeature.defined_execution_errors["ServerNotLocked"], message=message)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class Constraint(ABC):
    @abstractmethod
    def validate(self, *args, **kwargs) -> bool:
        """Return True if the given values are valid, False otherwise"""
        pass

    @classmethod
    @abstractmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType):
        identifier = xpath_sila(fdl_node, "name()")
        if identifier == "AllowedTypes":
            from sila2.framework.constraints.allowed_types import AllowedTypes

            return AllowedTypes.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "ContentType":
            from sila2.framework.constraints.content_type import ContentType

            return ContentType.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "ElementCount":
            from sila2.framework.constraints.element_count import ElementCount

            return ElementCount.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "FullyQualifiedIdentifier":
            from sila2.framework.constraints.fully_qualified_identifier import FullyQualifiedIdentifier

            return FullyQualifiedIdentifier.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "Length":
            from sila2.framework.constraints.length import Length

            return Length.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MaximalElementCount":
            from sila2.framework.constraints.maximal_element_count import MaximalElementCount

            return MaximalElementCount.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MaximalExclusive":
            from sila2.framework.constraints.maximal_exclusive import MaximalExclusive

            return MaximalExclusive.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MaximalInclusive":
            from sila2.framework.constraints.maximal_inclusive import MaximalInclusive

            return MaximalInclusive.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MaximalLength":
            from sila2.framework.constraints.maximal_length import MaximalLength

            return MaximalLength.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MinimalElementCount":
            from sila2.framework.constraints.minimal_element_count import MinimalElementCount

            return MinimalElementCount.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MinimalExclusive":
            from sila2.framework.constraints.minimal_exclusive import MinimalExclusive

            return MinimalExclusive.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MinimalInclusive":
            from sila2.framework.constraints.minimal_inclusive import MinimalInclusive

            return MinimalInclusive.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "MinimalLength":
            from sila2.framework.constraints.minimal_length import MinimalLength

            return MinimalLength.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "Pattern":
            from sila2.framework.constraints.pattern import Pattern

            return Pattern.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "Schema":
            from sila2.framework.constraints.schema import Schema

            return Schema.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "Set":
            from sila2.framework.constraints.set import Set

            return Set.from_fdl_node(fdl_node, parent_feature, base_type)
        if identifier == "Unit":
            from sila2.framework.constraints.unit import Unit

            return Unit.from_fdl_node(fdl_node, parent_feature, base_type)
        raise RuntimeError(f"Unknown constraint node: {identifier}")



from __future__ import annotations

import json
import os
from enum import Enum
from os.path import join
from typing import TYPE_CHECKING, Any, Union
from urllib.request import urlopen

import jsonschema
import xmlschema
from lxml import etree
from lxml.etree import XMLParser
from xmlschema import XMLSchemaValidationError

from sila2 import resource_dir
from sila2.framework.abc.constraint import Constraint
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class Schema(Constraint):
    schema_type: SchemaType
    source_type: SchemaSourceType
    source: str
    schema: Union[Any, xmlschema.XMLSchema, etree.XMLSchema]

    def __init__(self, schema_type: SchemaType, source_type: SchemaSourceType, source: str):
        self.schema_type = schema_type
        self.source_type = source_type
        self.source = source
        self.schema = self.get_schema()

    def validate(self, content: Union[str, bytes]) -> bool:
        if isinstance(self.schema, etree.XMLSchema):
            if isinstance(content, str):
                content = content.encode("utf-8")
            try:
                etree.fromstring(content, parser=XMLParser(schema=self.schema))
                return True
            except etree.XMLSyntaxError:
                return False
        elif isinstance(self.schema, xmlschema.XMLSchema):
            try:
                if isinstance(content, bytes):
                    content = content.decode("utf-8")
                self.schema.validate(content)
                return True
            except XMLSchemaValidationError:
                return False
        else:
            try:
                if isinstance(content, bytes):
                    content = content.decode("utf-8")
                jsonschema.validate(json.loads(content), schema=self.schema)
                return True
            except jsonschema.ValidationError:
                return False

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> Schema:
        _type = getattr(SchemaType, xpath_sila(fdl_node, "sila:Type/text()")[0])
        schema_node = xpath_sila(fdl_node, "sila:Inline|sila:Url")[0]
        source_type = getattr(SchemaSourceType, schema_node.xpath("name()"))
        schema_value = schema_node.text
        return cls(_type, source_type, schema_value)

    def get_schema(self) -> Union[Any, xmlschema.XMLSchema]:
        # XML
        if self.schema_type == SchemaType.Xml:
            if self.source_type == SchemaSourceType.Url:
                if "gitlab.com/sila2/sila_base" in self.source.lower() and self.source.split("/")[-1] in os.listdir(
                    join(resource_dir, "xsd")
                ):
                    return etree.XMLSchema(etree.parse(join(resource_dir, "xsd", self.source.split("/")[-1])))
                return xmlschema.XMLSchema(self.source)
            else:
                return etree.XMLSchema(etree.fromstring(self.source.encode("utf-8")))

        # JSON
        if self.source_type == SchemaSourceType.Inline:
            schema_str = self.source
        else:
            schema_str = urlopen(self.source).read().decode("utf-8")

        schema = json.loads(schema_str)
        validator_class = jsonschema.validators.validator_for(schema)
        validator_class.check_schema(schema)
        return schema

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.schema_type.name}, {self.source_type.name}, {self.source!r})"


class SchemaType(Enum):
    Xml = 0
    Json = 1


class SchemaSourceType(Enum):
    Inline = 0
    Url = 1



import os
import sys
from os.path import exists
from subprocess import PIPE, Popen
from typing import List

from sila2.code_generator.template_environment import TemplateEnvironment


class CodeGeneratorBase:
    overwrite: bool
    template_env: TemplateEnvironment
    generated_files: List[str]

    def __init__(self, overwrite: bool = False) -> None:
        self.overwrite = overwrite
        self.template_env = TemplateEnvironment()
        self.generated_files = []

    def generate_file(self, out_filename: str, content: str, *, allow_overwrite: bool = False) -> None:
        if exists(out_filename) and open(out_filename).read() != content and not (self.overwrite or allow_overwrite):
            raise FileExistsError(f"File '{out_filename}' already exists. Set --overwrite to overwrite existing files.")
        with open(out_filename, "w") as fp:
            fp.write(content)

        self.generated_files.append(out_filename)

    def format_generated_files(self) -> None:
        files = [file for file in self.generated_files if file.endswith((".py", ".pyi"))]
        proc_kwargs = dict(encoding="UTF-8", stdout=PIPE, stderr=PIPE)

        isort_proc = Popen(
            [sys.executable] + "-m isort --line-length 120 --quiet --profile black".split() + files,
            **proc_kwargs,
        )
        if isort_proc.wait() != 0:
            raise RuntimeError(f"Failed to run isort: {isort_proc.stderr.read()}")

        black_proc = Popen(
            [sys.executable] + "-m black --line-length 120 --quiet --fast".split() + files, **proc_kwargs
        )
        if black_proc.wait() != 0:
            raise RuntimeError(f"Failed to run black: {black_proc.stderr.read()}")

    def generate_directory(self, dir_: str, *, allow_overwrite: bool = False) -> None:
        if exists(dir_) and os.listdir(dir_) and not (self.overwrite or allow_overwrite):
            raise FileExistsError(f"Directory '{dir_}' already exists. Set --overwrite to overwrite existing files.")
        os.makedirs(dir_, exist_ok=True)



from typing import Optional

from sila2.framework.utils import xpath_sila


class NamedNode:
    _identifier: str
    _display_name: str
    _description: str

    def __init__(self, fdl_node):
        self._identifier = xpath_sila(fdl_node, "sila:Identifier")[0].text.strip()
        display_name: Optional[str] = xpath_sila(fdl_node, "sila:DisplayName")[0].text
        self._display_name = "" if display_name is None else display_name
        description: Optional[str] = xpath_sila(fdl_node, "sila:Description")[0].text
        self._description = "" if description is None else description



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

from sila2.features.lockcontroller import (
    InvalidLockIdentifier,
    LockControllerBase,
    LockControllerFeature,
    LockServer_Responses,
    ServerAlreadyLocked,
    ServerNotLocked,
    UnlockServer_Responses,
)
from sila2.framework import Command, Feature, FullyQualifiedIdentifier, Property
from sila2.server import MetadataInterceptor, SilaServer


@dataclass
class Lock:
    token: str
    timeout_duration: timedelta
    last_usage: datetime = field(default_factory=lambda: datetime.now())

    @property
    def is_expired(self):
        return (datetime.now() - self.timeout_duration) > self.last_usage


class LockControllerImpl(LockControllerBase):
    def __init__(self, parent_server: SilaServer):
        self.parent_server = parent_server
        self.lock: Optional[Lock] = None
        self.parent_server.add_metadata_interceptor(LockControllerInterceptor(self))

    @property
    def is_locked(self) -> bool:
        if self.lock is None:
            return False

        if self.lock.is_expired:
            self.lock = None

        return self.lock is not None

    def get_IsLocked(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> bool:
        return self.is_locked

    def LockServer(
        self, LockIdentifier: str, Timeout: int, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> LockServer_Responses:
        if self.is_locked:
            raise ServerAlreadyLocked

        self.lock = Lock(LockIdentifier, timedelta(seconds=Timeout))

    def UnlockServer(
        self, LockIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> UnlockServer_Responses:
        if not self.is_locked:
            raise ServerNotLocked

        if self.lock.token != LockIdentifier:
            raise InvalidLockIdentifier

        self.lock = None

    def get_calls_affected_by_LockIdentifier(self) -> List[Union[Feature, Command, Property, FullyQualifiedIdentifier]]:
        if not self.is_locked:
            return []

        # lock all other features
        result = []
        for feature in self.parent_server.features.values():
            if feature._identifier not in ["SiLAService", "LockController"]:
                result.append(feature)
        return result


class LockControllerInterceptor(MetadataInterceptor):
    def __init__(self, lockcontroller_impl: LockControllerImpl):
        self.lockidentifier_fqi = LockControllerFeature["LockIdentifier"].fully_qualified_identifier
        super().__init__([self.lockidentifier_fqi])
        self.lockcontroller_impl = lockcontroller_impl

    def intercept(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target_call: FullyQualifiedIdentifier
    ) -> None:
        token: str = metadata.pop(self.lockidentifier_fqi)
        if token != self.lockcontroller_impl.lock.token:
            raise InvalidLockIdentifier



from collections import namedtuple
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, Type, TypeVar

from google.protobuf.message import Message

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.intermediate_response import IntermediateResponse
from sila2.framework.command.parameter import Parameter
from sila2.framework.command.response import Response
from sila2.framework.errors.validation_error import ValidationError
from sila2.framework.property.property import Property

T = TypeVar("T", bound=NamedDataNode)


class CompositeMessageMappable(MessageMappable, Generic[T]):
    name: str
    fields: List[T]

    def __init__(self, fields: Iterable[T], message_type: Type[Message]):
        self.fields = list(fields)
        self.message_type = message_type
        self.name = message_type.__name__
        self.native_type = namedtuple(self.name, [field._identifier for field in fields])

    def __bool__(self):
        return bool(self.fields)

    def __iter__(self):
        return iter(self.fields)

    def to_message(self, *args, **kwargs) -> Message:
        # TODO: error messages
        #   - if multiple args are expected and only one integer is given: "int object has no length"
        #   - include fields names
        toplevel_named_data_node = (
            kwargs.pop("toplevel_named_data_node") if "toplevel_named_data_node" in kwargs else None
        )

        n_args = len(args) + len(kwargs)
        if len(self) == 0:
            if n_args == 0:
                return self.message_type()
            elif len(args) == 1 and not kwargs and args[0] is None:
                return self.message_type()
            else:
                raise TypeError(f"Expected no arguments, got {args}")

        # args[0] is NamedTuple representation of the message type
        if (
            len(args) == 1
            and not kwargs
            and isinstance(args[0], tuple)
            and hasattr(args[0], "_fields")
            and set(args[0]._fields) == set(f._identifier for f in self.fields)
        ):
            kwargs = {f._identifier: getattr(args[0], f._identifier) for f in self.fields}
            args = ()

        if len(args) == 1 and not kwargs and len(self) != 1:
            if isinstance(args[0], dict):
                kwargs = args[0]
                args = ()
            else:
                args = args[0]

        n_args = len(args) + len(kwargs)
        if n_args != len(self):
            raise TypeError(f"Message {self.message_type.__name__} has {len(self)} field(s), got {n_args} argument(s)")

        expected_kwargs_keys = {f._identifier for f in self.fields[len(args) :]}
        provided_kwargs_keys = set(kwargs.keys())
        if not expected_kwargs_keys.issubset(provided_kwargs_keys):
            raise TypeError(
                f"Message {self.message_type.__name__}: "
                f"Missing arguments for fields {expected_kwargs_keys - provided_kwargs_keys}"
            )

        field_values: Dict[str, Tuple[NamedDataNode, Any]] = {
            f._identifier: (f, arg) for f, arg in zip(self.fields[: len(args)], args)
        }
        for field in self.fields[len(args) :]:
            field_values[field._identifier] = field, kwargs[field._identifier]

        return self.message_type(
            **{
                field_id: field.data_type.to_message(
                    arg,
                    toplevel_named_data_node=field
                    if isinstance(field, (Parameter, Property, Response, IntermediateResponse))
                    else toplevel_named_data_node,
                )
                for field_id, (field, arg) in field_values.items()
            }
        )

    def to_native_type(self, message: Message, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Any:
        field_values = []
        for field in self.fields:
            value_msg = getattr(message, field._identifier)
            if isinstance(field, (Parameter, Property, Response, IntermediateResponse)):
                field_toplevel_named_data_node = field
            else:
                field_toplevel_named_data_node = toplevel_named_data_node

            try:
                field_values.append(field.data_type.to_native_type(value_msg, field_toplevel_named_data_node))
            except Exception as ex:
                if isinstance(ex, ValidationError):
                    raise ex
                if isinstance(field, Parameter):
                    raise ValidationError(field, f"{ex.__class__.__name__}: {ex}")
                raise ex

        return self.native_type(*field_values)

    def __len__(self):
        return len(self.fields)



from __future__ import annotations

import itertools as it
from os.path import join
from typing import List

from sila2.code_generator.code_generator_base import CodeGeneratorBase
from sila2.code_generator.template_objects.base import (
    ServerMetadata,
    ServerObservableCommand,
    ServerObservableProperty,
    ServerUnobservableCommand,
    ServerUnobservableProperty,
)
from sila2.code_generator.template_objects.basics import Field, Type
from sila2.code_generator.template_objects.client import (
    ClientMetadata,
    ClientObservableCommand,
    ClientObservableProperty,
    ClientUnobservableCommand,
    ClientUnobservableProperty,
)
from sila2.code_generator.template_objects.types import CompositeType, CompositeTypeField, TypeDefinition
from sila2.framework.data_types.data_type_definition import DataTypeDefinition
from sila2.framework.feature import Feature
from sila2.framework.utils import prettify_xml_string


class FeatureGenerator(CodeGeneratorBase):
    feature: Feature

    def __init__(self, feature: Feature, overwrite: bool) -> None:
        super().__init__(overwrite=overwrite)
        self.feature = feature
        self.feature_id_lowercase = feature._identifier.lower()

    def generate_feature_files(self, out_dir: str) -> None:
        with open(join(out_dir, f"{self.feature._identifier}.sila.xml"), "w") as fp:
            fp.write(prettify_xml_string(self.feature._feature_definition))

        self.generate_feature(out_dir)
        self.generate_client(out_dir)
        self.generate_base(out_dir)
        self.generate_init(out_dir)
        self.generate_errors(out_dir)
        self.generate_types(out_dir)

    def generate_errors(self, out_dir: str) -> None:
        content = self.template_env.get_template("feature/errors").render(
            feature=self.feature, errors=self.feature.defined_execution_errors.values()
        )
        self.generate_file(join(out_dir, f"{self.feature_id_lowercase}_errors.py"), content, allow_overwrite=True)

    def generate_init(self, out_dir: str) -> None:
        content = self.template_env.get_template("feature/init").render(feature=self.feature)
        self.generate_file(join(out_dir, "__init__.py"), content, allow_overwrite=True)

    def generate_feature(self, out_dir: str) -> None:
        content = self.template_env.get_template("feature/feature").render(feature=self.feature)
        self.generate_file(join(out_dir, f"{self.feature_id_lowercase}_feature.py"), content, allow_overwrite=True)

    def generate_types(self, out_dir: str) -> None:
        message_types: List[CompositeType] = []
        for cmd in it.chain(
            self.feature._unobservable_commands.values(),
            self.feature._observable_commands.values(),
        ):
            message_types.append(
                CompositeType(
                    f"{cmd._identifier}_Responses", [CompositeTypeField.from_field(f) for f in cmd.responses.fields]
                )
            )
        for cmd in self.feature._observable_commands.values():
            if cmd.intermediate_responses:
                message_types.append(
                    CompositeType(
                        f"{cmd._identifier}_IntermediateResponses",
                        [CompositeTypeField.from_field(f) for f in cmd.intermediate_responses.fields],
                    )
                )

        definitions = [
            TypeDefinition(t._identifier, Type.from_data_type(t.data_type))
            for t in self.feature._data_type_definitions.values()
        ]

        content = self.template_env.get_template("feature/types").render(
            named_tuples=message_types,
            data_type_definitions=definitions,
        )
        self.generate_file(join(out_dir, f"{self.feature_id_lowercase}_types.py"), content, allow_overwrite=True)

    def __generate_base_or_impl_str(self, generate_impl: bool) -> str:
        metadata = [ServerMetadata(m._identifier, m._description) for m in self.feature.metadata_definitions.values()]
        unobservable_properties = [
            ServerUnobservableProperty(p._identifier, Type.from_data_type(p.data_type), p._description)
            for p in self.feature._unobservable_properties.values()
        ]
        observable_properties = [
            ServerObservableProperty(p._identifier, Type.from_data_type(p.data_type), p._description)
            for p in self.feature._observable_properties.values()
        ]
        unobservable_commands = [
            ServerUnobservableCommand(
                cmd._identifier,
                [Field(p._identifier, Type.from_data_type(p.data_type), p._description) for p in cmd.parameters],
                [Field(r._identifier, Type.from_data_type(r.data_type), r._description) for r in cmd.responses],
                cmd._description,
            )
            for cmd in self.feature._unobservable_commands.values()
        ]
        observable_commands = [
            ServerObservableCommand(
                cmd._identifier,
                [Field(p._identifier, Type.from_data_type(p.data_type), p._description) for p in cmd.parameters],
                [
                    Field(i._identifier, Type.from_data_type(i.data_type), i._description)
                    for i in cmd.intermediate_responses
                ]
                if cmd.intermediate_responses is not None
                else [],
                [Field(r._identifier, Type.from_data_type(r.data_type), r._description) for r in cmd.responses],
                cmd._description,
            )
            for cmd in self.feature._observable_commands.values()
        ]
        imports = []
        for obj in it.chain(unobservable_properties, observable_properties, unobservable_commands, observable_commands):
            if isinstance(obj, ServerObservableProperty) and generate_impl:
                continue
            imports.extend(obj.imports)

        definition_imports = []
        for cmd in it.chain(self.feature._unobservable_commands.values(), self.feature._observable_commands.values()):
            for param in cmd.parameters:
                if isinstance(param.data_type, DataTypeDefinition):
                    definition_imports.append(param.data_type._identifier)

        return self.template_env.get_template("feature/impl" if generate_impl else "feature/base").render(
            feature=self.feature,
            imports=imports,
            metadata=metadata,
            unobservable_properties=unobservable_properties,
            observable_properties=observable_properties,
            unobservable_commands=unobservable_commands,
            observable_commands=observable_commands,
            definition_imports=definition_imports,
        )

    def generate_base(self, out_dir: str) -> None:
        content = self.__generate_base_or_impl_str(generate_impl=False)
        self.generate_file(join(out_dir, f"{self.feature_id_lowercase}_base.py"), content, allow_overwrite=True)

    def generate_impl(self, out_dir: str, *, prefix: str = "") -> None:
        content = self.__generate_base_or_impl_str(generate_impl=True)
        self.generate_file(join(out_dir, f"{prefix}{self.feature_id_lowercase}_impl.py"), content)

    def generate_client(self, out_dir: str) -> None:
        metadata = [
            ClientMetadata(m._identifier, m._description, Type.from_data_type(m.data_type))
            for m in self.feature.metadata_definitions.values()
        ]

        unobservable_properties = [
            ClientUnobservableProperty(prop._identifier, Type.from_data_type(prop.data_type), prop._description)
            for prop in self.feature._unobservable_properties.values()
        ]
        observable_properties = [
            ClientObservableProperty(prop._identifier, Type.from_data_type(prop.data_type), prop._description)
            for prop in self.feature._observable_properties.values()
        ]
        unobservable_commands = [
            ClientUnobservableCommand(
                cmd._identifier,
                [
                    Field(par._identifier, Type.from_data_type(par.data_type), par._description)
                    for par in cmd.parameters
                ],
                cmd._description,
            )
            for cmd in self.feature._unobservable_commands.values()
        ]
        observable_commands = [
            ClientObservableCommand(
                cmd._identifier,
                [
                    Field(par._identifier, Type.from_data_type(par.data_type), par._description)
                    for par in cmd.parameters
                ],
                cmd._description,
                bool(cmd.intermediate_responses),
            )
            for cmd in self.feature._observable_commands.values()
        ]

        imports = []
        for obj in it.chain(
            unobservable_properties, observable_properties, unobservable_commands, observable_commands, metadata
        ):
            imports.extend(obj.imports)

        definition_imports = []
        for cmd in it.chain(self.feature._unobservable_commands.values(), self.feature._observable_commands.values()):
            for param in cmd.parameters:
                if isinstance(param.data_type, DataTypeDefinition):
                    definition_imports.append(param.data_type._identifier)

        content = self.template_env.get_template("feature/client").render(
            feature=self.feature,
            imports=imports,
            metadata=metadata,
            unobservable_properties=unobservable_properties,
            observable_properties=observable_properties,
            unobservable_commands=unobservable_commands,
            observable_commands=observable_commands,
            definition_imports=definition_imports,
        )
        self.generate_file(join(out_dir, f"{self.feature_id_lowercase}_client.pyi"), content, allow_overwrite=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.utils import call_rpc_function
from sila2.framework.property.unobservable_property import UnobservableProperty

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature
    from sila2.client.client_metadata import ClientMetadataInstance

T = TypeVar("T")


class ClientUnobservableProperty(Generic[T]):
    _parent_feature: ClientFeature
    _wrapped_property: UnobservableProperty

    def __init__(self, parent_feature: ClientFeature, wrapped_property: UnobservableProperty):
        self._parent_feature = parent_feature
        self._wrapped_property = wrapped_property

    def get(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> T:
        param_msg = self._wrapped_property.get_parameters_message()

        response_msg = call_rpc_function(
            getattr(self._parent_feature._grpc_stub, f"Get_{self._wrapped_property._identifier}"),
            param_msg,
            metadata=metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_property,
        )

        return self._wrapped_property.to_native_type(response_msg)



import time
from datetime import datetime
from typing import List, Optional, Union
from uuid import UUID

from zeroconf import ServiceBrowser, Zeroconf

from sila2.client.sila_client import SilaClient
from sila2.discovery.listener import SilaServiceListener


class SilaDiscoveryBrowser(ServiceBrowser):
    listener: SilaServiceListener
    clients: List[SilaClient]

    def __init__(
        self,
        *,
        insecure: bool = False,
        root_certs: Optional[bytes] = None,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
    ):
        self.listener = SilaServiceListener(
            self, insecure=insecure, root_certs=root_certs, private_key=private_key, cert_chain=cert_chain
        )
        super().__init__(Zeroconf(), "_sila._tcp.local.", self.listener)

    @property
    def clients(self) -> List[SilaClient]:
        return list(self.listener.services.values())

    def find_server(
        self, server_name: Optional[str] = None, server_uuid: Optional[Union[UUID, str]] = None, timeout: float = 0
    ) -> SilaClient:
        if timeout < 0:
            raise ValueError("Timeout must be non-negative")

        start_time = datetime.now()
        while timeout == 0 or (datetime.now() - start_time).total_seconds() < timeout:
            for client in self.clients:
                if server_name is not None and client.SiLAService.ServerName.get() != server_name:
                    continue
                if server_uuid is not None and client.SiLAService.ServerUUID.get() != str(server_uuid):
                    continue
                return client
            time.sleep(0.1)

        raise TimeoutError(f"No suitable SiLA server was found after {timeout} seconds")

    def __del__(self):
        self.cancel()



from sila2.framework.binary_transfer.binary_transfer_error import BinaryTransferError, BinaryTransferErrorType


class BinaryDownloadFailed(BinaryTransferError):
    def __init__(self, message: str):
        super().__init__(BinaryTransferErrorType.BINARY_DOWNLOAD_FAILED, message)



from __future__ import annotations

from typing import List, NamedTuple

from sila2.code_generator.template_objects.basics import Field, Import, Type


class ClientUnobservableProperty(NamedTuple):
    name: str
    type: Type
    docstring: str

    @property
    def imports(self) -> List[Import]:
        return self.type.imports


class ClientObservableProperty(NamedTuple):
    name: str
    type: Type
    docstring: str

    @property
    def imports(self) -> List[Import]:
        return self.type.imports


class ClientUnobservableCommand(NamedTuple):
    name: str
    parameters: List[Field]
    docstring: str

    @property
    def imports(self) -> List[Import]:
        ret = []
        for par in self.parameters:
            ret.extend(par.imports)
        return ret


class ClientObservableCommand(NamedTuple):
    name: str
    parameters: List[Field]
    docstring: str
    has_intermediate_responses: bool

    @property
    def imports(self) -> List[Import]:
        ret = []
        for par in self.parameters:
            ret.extend(par.imports)
        return ret


class ClientMetadata(NamedTuple):
    name: str
    docstring: str
    type: Type

    @property
    def imports(self) -> List[Import]:
        return self.type.imports



from __future__ import annotations

import warnings
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic, Optional, TypeVar
from uuid import UUID

from sila2.client.execution_info_subscription_thread import ExecutionInfoSubscriptionThread
from sila2.client.no_intermediate_responses import NoIntermediateResponses
from sila2.client.subscription_stream import SubscriptionStream
from sila2.client.utils import call_rpc_function
from sila2.framework.command.execution_info import CommandExecutionInfo, CommandExecutionStatus

if TYPE_CHECKING:
    from sila2.client.client_observable_command import ClientObservableCommand
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandExecutionUUID as SilaCommandExecutionUUID

ResultType = TypeVar("ResultType")
IntermediateType = TypeVar("IntermediateType")


class ClientObservableCommandInstance(Generic[IntermediateType, ResultType]):
    _client_command: ClientObservableCommand
    execution_uuid: UUID
    status: Optional[CommandExecutionStatus]
    progress: Optional[float]
    estimated_remaining_time: Optional[timedelta]
    lifetime_of_execution: Optional[timedelta]
    done: bool
    __last_lifetime_update_timestamp: datetime
    __last_lifetime_update_duration: Optional[timedelta]
    __last_remaining_time_update_timestamp: datetime
    __last_remaining_time_update_duration: Optional[timedelta]
    __info_update_thread: ExecutionInfoSubscriptionThread

    def __init__(
        self,
        client_command: ClientObservableCommand,
        execution_uuid: UUID,
        lifetime_of_execution: Optional[timedelta] = None,
    ):
        self._client_command = client_command
        self.execution_uuid = execution_uuid
        self.__last_remaining_time_update_duration = None
        self.__last_remaining_time_update_timestamp = datetime.now()
        self.__last_lifetime_update_duration = None
        self.__last_lifetime_update_timestamp = datetime.now()
        self.lifetime_of_execution = lifetime_of_execution
        self.status = None
        self.progress = None
        self.estimated_remaining_time = None
        self.done = False
        self.__info_update_thread = ExecutionInfoSubscriptionThread(
            self._client_command._wrapped_command.fully_qualified_identifier,
            self.execution_uuid,
            getattr(client_command._parent_feature._grpc_stub, f"{client_command._wrapped_command._identifier}_Info")(
                self.__get_execution_uuid_message()
            ),
            self,
        )
        self.__info_update_thread.start()

    @property
    def lifetime_of_execution(self) -> Optional[timedelta]:
        if self.__last_lifetime_update_duration is None:
            return None
        lifetime = datetime.now() - self.__last_lifetime_update_timestamp + self.__last_lifetime_update_duration
        if lifetime.total_seconds() < 0:
            return timedelta(0)
        return lifetime

    @lifetime_of_execution.setter
    def lifetime_of_execution(self, value: Optional[timedelta]) -> None:
        if self.lifetime_of_execution is None:
            if value is not None:
                warnings.warn(
                    f"Server shortened previously unconstrained lifetime of observable command instance "
                    f"{self.execution_uuid}. Ignoring."
                )
            return
        if value is None:
            warnings.warn("Server did not provide an updated lifetime of execution. Keeping the previous one.")
            return
        if value < self.lifetime_of_execution:
            warnings.warn(
                f"Server shortened the lifetime of observable command {self.execution_uuid} from "
                f"{self.lifetime_of_execution} to {value}. Ignoring."
            )
            return
        self.__last_lifetime_update_duration = value
        self.__last_lifetime_update_timestamp = datetime.now()

    @property
    def estimated_remaining_time(self) -> Optional[timedelta]:
        if self.__last_remaining_time_update_duration is None:
            return None
        remaining = self.__last_remaining_time_update_duration - (
            datetime.now() - self.__last_remaining_time_update_timestamp
        )
        if remaining.total_seconds() < 0:
            return timedelta(0)
        return remaining

    @estimated_remaining_time.setter
    def estimated_remaining_time(self, value: Optional[timedelta]) -> None:
        if value is None:
            return
        self.__last_remaining_time_update_timestamp = datetime.now()
        self.__last_remaining_time_update_duration = value

    def _update(self, new_execution_info: CommandExecutionInfo) -> None:
        self.status = new_execution_info.status
        self.progress = new_execution_info.progress
        self.estimated_remaining_time = new_execution_info.estimated_remaining_time
        self.lifetime_of_execution = new_execution_info.updated_lifetime_of_execution
        if new_execution_info.status in (
            CommandExecutionStatus.finishedSuccessfully,
            CommandExecutionStatus.finishedWithError,
        ):
            self.done = True

    def subscribe_intermediate_responses(self) -> SubscriptionStream[IntermediateType]:
        if self._client_command._wrapped_command.intermediate_responses is None:
            raise NoIntermediateResponses(
                f"Command {self._client_command._wrapped_command._identifier} has no intermediate responses"
            )

        rpc_func = getattr(
            self._client_command._parent_feature._grpc_stub,
            f"{self._client_command._wrapped_command._identifier}_Intermediate",
        )
        return SubscriptionStream(
            rpc_func(self.__get_execution_uuid_message()),
            self._client_command._wrapped_command.intermediate_responses.to_native_type,
        )

    def get_responses(self) -> ResultType:
        rpc_func = getattr(
            self._client_command._parent_feature._grpc_stub,
            f"{self._client_command._wrapped_command._identifier}_Result",
        )
        response_msg = call_rpc_function(
            rpc_func,
            self.__get_execution_uuid_message(),
            metadata=None,
            client=self._client_command._parent_feature._parent_client,
            origin=self._client_command._wrapped_command,
        )
        # response_msg = rpc_func(self.__get_execution_uuid_message())
        return self._client_command._wrapped_command.responses.to_native_type(response_msg)

    def __get_execution_uuid_message(self) -> SilaCommandExecutionUUID:
        return self._client_command._parent_feature._pb2_module.SiLAFramework__pb2.CommandExecutionUUID(
            value=str(self.execution_uuid)
        )



from __future__ import annotations

import logging
import re
from collections import defaultdict
from contextlib import contextmanager
from queue import Queue
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, List, NoReturn, Optional, Set, Tuple, Union
from uuid import UUID

import grpc
from google.protobuf.message import Message
from grpc import ServicerContext

from sila2.framework import (
    Command,
    DefinedExecutionError,
    InvalidMetadata,
    Metadata,
    NoMetadataAllowed,
    Property,
    UndefinedExecutionError,
    ValidationError,
)
from sila2.framework.abc.sila_error import SilaError
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.errors.invalid_command_execution_uuid import InvalidCommandExecutionUUID
from sila2.framework.feature import Feature
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.property.observable_property import ObservableProperty
from sila2.framework.utils import FullyQualifiedIdentifierRegex, raise_as_rpc_error
from sila2.server.feature_implementation_base import FeatureImplementationBase
from sila2.server.observables.observable_command_manager import ObservableCommandManager
from sila2.server.observables.subscription_manager_thread import SubscriptionManagerThread

if TYPE_CHECKING:
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandExecutionUUID as SilaCommandExecutionUUID
    from sila2.server import ObservableCommandInstance
    from sila2.server.sila_server import SilaServer


class FeatureImplementationServicer:
    parent_server: SilaServer
    feature: Feature
    implementation: Optional[FeatureImplementationBase]
    observable_command_managers: Dict[str, Dict[UUID, ObservableCommandManager]]
    observable_property_subscription_managers: Dict[str, Dict[Optional[Queue], SubscriptionManagerThread]]

    def __init__(self, parent_server: SilaServer, feature: Feature):
        self.parent_server = parent_server
        self.feature = feature
        self.implementation = None
        self.observable_command_managers = defaultdict(dict)
        self.observable_property_subscription_managers = {}
        self.logger = logging.getLogger(feature._identifier)

    def _set_implementation(self, implementation: FeatureImplementationBase):
        self.logger.debug(f"Setting implementation to {implementation}")
        self.implementation = implementation
        self.__start_observable_property_listeners()

        for prop in self.feature._unobservable_properties.values():
            setattr(self, f"Get_{prop._identifier}", self.__get_unobservable_property_get_call(prop._identifier))
        for cmd in self.feature._unobservable_commands.values():
            setattr(self, cmd._identifier, self.__get_unobservable_command_init_call(cmd._identifier))
        for prop in self.feature._observable_properties.values():
            setattr(
                self, f"Subscribe_{prop._identifier}", self.__get_observable_property_subscribe_call(prop._identifier)
            )
        for cmd in self.feature._observable_commands.values():
            setattr(self, cmd._identifier, self.__get_observable_command_init_call(cmd._identifier))
            setattr(self, f"{cmd._identifier}_Info", self.__get_observable_command_info_subscribe_call(cmd._identifier))
            setattr(
                self,
                f"{cmd._identifier}_Intermediate",
                self.__get_observable_command_intermediate_subscribe_call(cmd._identifier),
            )
            setattr(self, f"{cmd._identifier}_Result", self.__get_observable_command_result_get_call(cmd._identifier))
        for metadata in self.feature.metadata_definitions.values():
            setattr(
                self,
                f"Get_FCPAffectedByMetadata_{metadata._identifier}",
                self.__get_fpc_affected_by_metadata_call(metadata._identifier),
            )

    def __start_observable_property_listeners(self):
        self.logger.debug("Starting subscription managers for observable properties")
        for prop in self.feature._observable_properties.values():
            manager = SubscriptionManagerThread(
                prop.fully_qualified_identifier,
                getattr(self.implementation, f"_{prop._identifier}_producer_queue"),
                prop.to_message,
            )
            self.observable_property_subscription_managers[prop._identifier] = {None: manager}
            manager.start()

    def __get_observable_command_manager(
        self, command_id: str, execution_uuid: UUID, context: ServicerContext
    ) -> ObservableCommandManager:
        manager = self.observable_command_managers[command_id].get(execution_uuid)
        if manager is None:
            self.__raise_as_rpc_error(
                InvalidCommandExecutionUUID(f"No running command instance with the execution uuid {execution_uuid}"),
                context,
            )
        return manager

    def __get_unobservable_command_init_call(self, command_id: str) -> Callable[[Message, ServicerContext], Message]:
        cmd = self.feature._unobservable_commands[command_id]
        impl_func: Callable = getattr(self.implementation, command_id)

        def wrapper(request: Message, context: ServicerContext) -> Message:
            self.logger.info(f"Request: unobservable command {command_id}")
            metadata = self.__extract_metadata(context, self.parent_server, cmd)
            allowed_errors = self.__find_allowed_defined_execution_errors(self.parent_server, cmd, metadata.keys())
            params = self.__unpack_parameters(cmd, request, context)

            with raises_rpc_errors(context, self.logger, allowed_errors):
                self.__apply_metadata_interceptors(None, metadata, cmd.fully_qualified_identifier)
                response = impl_func(*params, metadata=metadata)
                self.logger.info(f"Implementation returned {response}")
                return cmd.responses.to_message(response)

        return wrapper

    def __get_unobservable_property_get_call(self, property_id: str) -> Callable[[Message, ServicerContext], Message]:
        prop = self.feature._unobservable_properties[property_id]
        impl_func: Callable = getattr(self.implementation, f"get_{property_id}")

        def wrapper(request, context: ServicerContext):
            self.logger.info(f"Request: unobservable property {property_id}")
            metadata = self.__extract_metadata(context, self.parent_server, prop)
            allowed_errors = self.__find_allowed_defined_execution_errors(self.parent_server, prop, metadata.keys())

            with raises_rpc_errors(context, self.logger, allowed_errors):
                self.__apply_metadata_interceptors(None, metadata, prop.fully_qualified_identifier)
                response = impl_func(metadata=metadata)
                self.logger.info(f"Implementation returned {response}")
                return prop.to_message(response)

        return wrapper

    def __get_observable_command_init_call(self, command_id: str) -> Callable[[Message, ServicerContext], Message]:
        cmd = self.feature._observable_commands[command_id]
        impl_func: Callable = getattr(self.implementation, command_id)

        def wrapper(request: Message, context: ServicerContext) -> Message:
            self.logger.info(f"Request: observable command initiation for {command_id}")
            metadata = self.__extract_metadata(context, self.parent_server, cmd)

            params = self.__unpack_parameters(cmd, request, context)

            def _func_to_execute(instance: ObservableCommandInstance) -> Message:
                self.__apply_metadata_interceptors(None, metadata, cmd.fully_qualified_identifier)
                response = impl_func(
                    *params,
                    metadata=metadata,
                    instance=instance,
                )
                self.logger.info(f"Implementation returned {response}")
                return cmd.responses.to_message(response)

            command_manager = ObservableCommandManager(self.parent_server, cmd, _func_to_execute, metadata.keys())
            self.observable_command_managers[command_id][command_manager.command_execution_uuid] = command_manager

            return self.feature._pb2_module.SiLAFramework__pb2.CommandConfirmation(
                commandExecutionUUID=self.feature._pb2_module.SiLAFramework__pb2.CommandExecutionUUID(
                    value=str(command_manager.command_execution_uuid)
                )
            )

        return wrapper

    def __get_observable_command_info_subscribe_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Iterator[Message]]:
        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Iterator[Message]:
            self.logger.info(f"Request: observable command info subscription for {command_id}")
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)

            with raises_rpc_errors(context, self.logger):
                for value in manager.subscribe_to_execution_infos(context):
                    yield value

        return wrapper

    def __get_observable_command_intermediate_subscribe_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Iterator[Message]]:
        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Iterator[Message]:
            self.logger.info(f"Request: observable command intermediate response subscription for {command_id}")
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)

            with raises_rpc_errors(context, self.logger):
                for value in manager.subscribe_to_intermediate_responses(context):
                    yield value

        return wrapper

    def __get_observable_command_result_get_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Message]:
        cmd = self.feature._observable_commands[command_id]

        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Message:
            self.logger.info(f"Request: observable command result for {command_id}")
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)
            allowed_errors = self.__find_allowed_defined_execution_errors(
                self.parent_server, cmd, manager.metadata_identifiers
            )

            with raises_rpc_errors(context, self.logger, allowed_errors):
                responses = manager.get_responses()
                self.logger.info(f"Returning {responses}")
                return responses

        return wrapper

    def __get_observable_property_subscribe_call(
        self, property_id: str
    ) -> Callable[[Message, ServicerContext], Iterator[Message]]:
        prop = self.feature._observable_properties[property_id]
        impl_func = getattr(self.implementation, f"{property_id}_on_subscription")

        def wrapper(request: Message, context: ServicerContext) -> Iterator[Message]:
            self.logger.info(f"Request: observable property subscription for {property_id}")
            metadata = self.__extract_metadata(context, self.parent_server, prop)
            allowed_errors = self.__find_allowed_defined_execution_errors(self.parent_server, prop, metadata.keys())

            with raises_rpc_errors(context, self.logger, allowed_errors):
                self.__apply_metadata_interceptors(None, metadata, prop.fully_qualified_identifier)

                queue = impl_func(metadata=metadata)
                manager = self.__get_observable_property_subscription_thread_for_queue(prop, queue)

                for value in manager.add_subscription(context):
                    yield value

        return wrapper

    def __get_fpc_affected_by_metadata_call(self, metadata_id: str) -> Callable[[Message, ServicerContext], Message]:
        metadata_node = self.feature.metadata_definitions[metadata_id]
        impl_func = getattr(self.implementation, f"get_calls_affected_by_{metadata_id}")

        def wrapper(request: Message, context: ServicerContext) -> Message:
            self.logger.info(f"Request: Calls affected by {metadata_id}")
            with raises_rpc_errors(context, self.logger):
                affected_calls = impl_func()
                self.logger.info(f"Returning {affected_calls}")
                return metadata_node.to_affected_calls_message(affected_calls)

        return wrapper

    def cancel_all_subscriptions(self):
        self.logger.info("Cancelling all subscriptions")
        for managers in self.observable_property_subscription_managers.values():
            for manager in managers.values():
                manager.cancel_producer()

    def __apply_metadata_interceptors(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target: FullyQualifiedIdentifier
    ):
        for interceptor in self.parent_server.metadata_interceptors:
            if interceptor.required_metadata.issubset(metadata.keys()):
                self.logger.info(f"Applying metadata interceptor {interceptor}")
                interceptor.intercept(parameters, metadata, target)

    def __unpack_parameters(self, command: Command, request, context: ServicerContext) -> Tuple[Any, ...]:
        try:
            parameters = command.parameters.to_native_type(request)
            self.logger.info(f"Unpacked parameters: {parameters}")
            return parameters
        except ValidationError as val_err:
            self.__raise_as_rpc_error(val_err, context)

    def __find_allowed_defined_execution_errors(
        self,
        server: SilaServer,
        origin: Optional[Union[Command, Property]] = None,
        metadata_identifiers: Optional[Iterable[FullyQualifiedIdentifier]] = None,
    ) -> List[DefinedExecutionErrorNode]:
        allowed_errors = []
        if origin is not None:
            allowed_errors.extend(origin.defined_execution_errors)
        if metadata_identifiers is not None:
            for metadata_id in metadata_identifiers:
                allowed_errors.extend(
                    server.children_by_fully_qualified_identifier[metadata_id].defined_execution_errors
                )
        self.logger.debug(f"Allowed defined execution errors: {[err._identifier for err in allowed_errors]}")
        return allowed_errors

    def __extract_metadata(
        self, context: grpc.ServicerContext, server: SilaServer, origin: Union[Property, Command]
    ) -> Dict[FullyQualifiedIdentifier, Any]:
        self.logger.debug("Extracting metadata")
        expected_metadata: Set[FullyQualifiedIdentifier] = set()
        for feature_servicer in server.feature_servicers.values():
            for meta_id, meta in feature_servicer.feature.metadata_definitions.items():
                raw_affected_calls: List[Union[Feature, Command, Property, FullyQualifiedIdentifier]] = getattr(
                    feature_servicer.implementation, f"get_calls_affected_by_{meta_id}"
                )()
                affected_fqis = [
                    obj if isinstance(obj, FullyQualifiedIdentifier) else obj.fully_qualified_identifier
                    for obj in raw_affected_calls
                ]

                if (
                    origin.fully_qualified_identifier in affected_fqis
                    or origin.parent_feature.fully_qualified_identifier in affected_fqis
                ):
                    expected_metadata.add(meta.fully_qualified_identifier)
        self.logger.debug(f"Required metadata: {expected_metadata}")

        received_metadata: Dict[FullyQualifiedIdentifier, Any] = {}
        for key, value in context.invocation_metadata():
            if re.fullmatch(
                f"sila/{FullyQualifiedIdentifierRegex.MetadataIdentifier}/bin",
                key.replace("-", "/"),
                flags=re.IGNORECASE,
            ):
                try:
                    key = key[5:-4].replace("-", "/")
                    meta: Metadata = server.children_by_fully_qualified_identifier[FullyQualifiedIdentifier(key)]
                    received_metadata[meta.fully_qualified_identifier] = meta.to_native_type(value)
                except KeyError:
                    self.__raise_as_rpc_error(InvalidMetadata(f"Server has no metadata {key}"), context)
                except Exception as ex:
                    self.__raise_as_rpc_error(
                        InvalidMetadata(f"Failed to deserialize metadata value for {key!r}: {ex}"), context
                    )
        self.logger.debug(f"Received metadata: {received_metadata}")

        # local import to prevent circular imports
        from sila2.features.silaservice import SiLAServiceFeature

        if origin.parent_feature.fully_qualified_identifier == SiLAServiceFeature.fully_qualified_identifier:
            if received_metadata:
                self.__raise_as_rpc_error(
                    NoMetadataAllowed("Cannot use metadata with calls to the SiLAService feature"), context
                )
            if expected_metadata:
                self.logger.warning(f"Server expects metadata for the SiLAService call {origin._identifier}. Ignoring.")
            return {}

        for expected_meta_fqi in expected_metadata:
            if expected_meta_fqi not in received_metadata:
                self.__raise_as_rpc_error(
                    InvalidMetadata(f"Did not receive required metadata {expected_meta_fqi}"), context
                )

        return received_metadata

    def __raise_as_rpc_error(self, error: SilaError, context: grpc.ServicerContext) -> NoReturn:
        self.logger.error(f"Raising as RpcError {error!r}")
        raise_as_rpc_error(error, context)

    def __get_observable_property_subscription_thread_for_queue(
        self, prop: ObservableProperty, queue: Optional[Queue]
    ) -> SubscriptionManagerThread:
        prop_managers = self.observable_property_subscription_managers[prop._identifier]

        # default queue
        if queue is None:
            return prop_managers[None]

        # subscribe to existing custom queue
        if queue in prop_managers:
            return prop_managers[queue]

        # new custom queue
        manager = SubscriptionManagerThread(prop.fully_qualified_identifier, queue, prop.to_message)
        prop_managers[queue] = manager
        manager.start()
        return manager


@contextmanager
def raises_rpc_errors(
    context: ServicerContext,
    logger: logging.Logger,
    allowed_defined_execution_errors: Optional[Iterable[DefinedExecutionErrorNode]] = None,
):
    try:
        yield
    except Exception as ex:
        logger.exception("Caught exception, raising as RpcError")
        if isinstance(ex, NotImplementedError):
            context.abort(grpc.StatusCode.UNIMPLEMENTED, "The requested functionality is not implemented")

        if not isinstance(ex, SilaError):
            raise_as_rpc_error(UndefinedExecutionError(ex), context)

        allowed_error_identifiers = [err.fully_qualified_identifier for err in allowed_defined_execution_errors]
        if isinstance(ex, DefinedExecutionError) and ex.fully_qualified_identifier not in allowed_error_identifiers:
            raise_as_rpc_error(UndefinedExecutionError(ex), context)

        raise_as_rpc_error(ex, context)



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.constraint import Constraint

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class ElementCount(Constraint):
    element_count: int

    def __init__(self, element_count: int):
        self.element_count = element_count

    def validate(self, value: list) -> bool:
        return len(value) == self.element_count

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> ElementCount:
        return cls(int(fdl_node.text))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.element_count})"



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from zeroconf import ServiceInfo

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer


class SilaServiceInfo(ServiceInfo):
    sila_server: SilaServer

    def __init__(self, server: SilaServer, address: str, port: int, ca: Optional[bytes] = None):
        properties = dict(
            version=server.server_version,
            server_name=server.server_name,
            description=server.server_description,
        )

        if ca is not None:
            current_ca_pos = 0
            current_line_num = 0
            while current_ca_pos < len(ca):
                key = f"ca{current_line_num}"
                current_line_num += 1

                value_length = 255 - len(key) - 1  # TXT records have max length 255 and are formatted as "key=value"
                properties[key] = ca[current_ca_pos : current_ca_pos + value_length]
                current_ca_pos += value_length

        super().__init__(
            type_="_sila._tcp.local.",
            name=f"{server.server_uuid}._sila._tcp.local.",
            parsed_addresses=[address],
            port=port,
            properties=properties,
        )
        self.sila_server = server



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Union

from sila2.client.client_metadata import ClientMetadata
from sila2.client.client_observable_command import ClientObservableCommand
from sila2.client.client_observable_property import ClientObservableProperty
from sila2.client.client_unobservable_command import ClientUnobservableCommand
from sila2.client.client_unobservable_property import ClientUnobservableProperty
from sila2.framework.binary_transfer.client_binary_transfer_handler import ClientBinaryTransferHandler
from sila2.framework.feature import Feature

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient

GrpcStub = Any


class ClientFeature(Feature):
    _grpc_stub: GrpcStub
    _parent_client: SilaClient
    _binary_transfer_handler: ClientBinaryTransferHandler
    _client_commands: Dict[str, Union[ClientUnobservableCommand, ClientObservableCommand]]
    _client_properties: Dict[str, Union[ClientUnobservableProperty, ClientObservableProperty]]
    _client_metadata: Dict[str, ClientMetadata]

    def __init__(self, feature_definition: str, parent_client: SilaClient):
        super().__init__(feature_definition)
        self._parent_client = parent_client
        self._grpc_stub = getattr(self._grpc_module, f"{self._identifier}Stub")(parent_client._channel)
        self._binary_transfer_handler = ClientBinaryTransferHandler(parent_client._channel)

        self._client_properties = {}
        self._client_commands = {}
        self._client_metadata = {}
        for identifier, prop in self._unobservable_properties.items():
            client_prop = ClientUnobservableProperty(self, prop)
            self._client_properties[identifier] = client_prop
            setattr(self, identifier, client_prop)
        for identifier, prop in self._observable_properties.items():
            client_prop = ClientObservableProperty(self, prop)
            self._client_properties[identifier] = client_prop
            setattr(self, identifier, ClientObservableProperty(self, prop))
        for identifier, cmd in self._unobservable_commands.items():
            client_cmd = ClientUnobservableCommand(self, cmd)
            self._client_commands[identifier] = client_cmd
            setattr(self, identifier, client_cmd)
        for identifier, cmd in self._observable_commands.items():
            client_cmd = ClientObservableCommand(self, cmd)
            self._client_commands[identifier] = client_cmd
            setattr(self, identifier, client_cmd)
        for identifier, meta in self.metadata_definitions.items():
            client_metadata = ClientMetadata(self, meta)
            self._client_metadata[identifier] = client_metadata
            setattr(self, identifier, client_metadata)



from __future__ import annotations

from typing import TYPE_CHECKING, List

from sila2.framework.abc.constraint import Constraint
from sila2.framework.data_types.any import SilaAnyType
from sila2.framework.utils import xml_node_to_normalized_string, xpath_sila

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class AllowedTypes(Constraint):
    allowed_types: List[str]

    def __init__(self, allowed_types: List[str]):
        self.allowed_types = allowed_types

    def validate(self, value: SilaAnyType) -> bool:
        _type_str = xml_node_to_normalized_string(value.type_xml, remove_namespace=True)
        return any(_type_str == s for s in self.allowed_types)

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> AllowedTypes:
        return cls(
            [
                xml_node_to_normalized_string(dtype_node, remove_namespace=True)
                for dtype_node in xpath_sila(fdl_node, "sila:DataType")
            ]
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.allowed_types!r})"



import os

from jinja2 import FileSystemLoader

from sila2 import resource_dir


class TemplateLoader(FileSystemLoader):
    def __init__(self):
        template_dir = os.path.join(resource_dir, "code_generator_templates")
        super().__init__(searchpath=template_dir, encoding="utf-8")

    def get_source(self, environment, template: str):
        return super().get_source(environment, f"{template}.jinja2")



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Dict, Optional, Type
from uuid import UUID

from grpc import Server

from sila2.framework.abc.binary_transfer_handler import BinaryTransferHandler
from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.binary_transfer.download_servicer import BinaryDownloadServicer
from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID
from sila2.framework.binary_transfer.upload_servicer import BinaryUploadServicer

if TYPE_CHECKING:
    from sila2.pb2_stubs.SiLAFramework_pb2 import Binary as SilaBinary


class ServerBinaryTransferHandler(BinaryTransferHandler):
    upload_servicer: binary_transfer_grpc_module.BinaryUploadServicer
    download_servicer: binary_transfer_grpc_module.BinaryDownloadServicer
    known_binaries: Dict[UUID, bytes]

    def __init__(self, grpc_server: Server):
        self.upload_servicer = BinaryUploadServicer(self)
        self.download_servicer = BinaryDownloadServicer(self)
        binary_transfer_grpc_module.add_BinaryUploadServicer_to_server(self.upload_servicer, grpc_server)
        binary_transfer_grpc_module.add_BinaryDownloadServicer_to_server(self.download_servicer, grpc_server)

        self.known_binaries = {}

    def to_native_type(self, binary_uuid: UUID, toplevel_named_data_node: Optional[NamedDataNode] = None) -> bytes:
        try:
            return self.known_binaries[binary_uuid]
        except KeyError:
            raise InvalidBinaryTransferUUID(f"Invalid binary transfer UUID: {binary_uuid}")

    def to_message(
        self, binary: bytes, message_type: Type[SilaBinary], toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> SilaBinary:
        binary_uuid = uuid.uuid4()
        self.known_binaries[binary_uuid] = binary
        return message_type(binaryTransferUUID=str(binary_uuid))



import re
from distutils.dir_util import copy_tree
from glob import glob
from os import makedirs
from os.path import basename, dirname, join
from shutil import copy

from setuptools import setup

base_dir = dirname(__file__)


def copy_resources_from_sila_base():
    resource_dir = join(base_dir, "src", "sila2", "resources")
    makedirs(resource_dir, exist_ok=True)

    # xsd
    copy_tree(join(base_dir, "sila_base", "schema"), join(resource_dir, "xsd"))

    # xsl
    makedirs(join(resource_dir, "xsl"), exist_ok=True)
    for file in glob(join(base_dir, "sila_base", "xslt", "*.xsl")):
        copy(file, join(resource_dir, "xsl", basename(file)))

    # proto
    makedirs(join(resource_dir, "proto"), exist_ok=True)
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLAFramework.proto"),
        join(resource_dir, "proto", "SiLAFramework.proto"),
    )
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLABinaryTransfer.proto"),
        join(resource_dir, "proto", "SiLABinaryTransfer.proto"),
    )


def prepare_readme():
    """README contains gitlab-internal links, which need to be extended for PyPI"""
    content = open(join(dirname(__file__), "README.md")).read()
    link_pattern = r"\]\(([^h][^)]+)\)"
    return re.sub(link_pattern, "](https://gitlab.com/sila2/sila_python/-/tree/master/\\1)", content)


copy_resources_from_sila_base()

setup(
    long_description=prepare_readme(),
)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Optional, Type, Union

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Real as SilaReal


class Real(DataType):
    native_type = float
    message_type: Type[SilaReal]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Real

    def to_native_type(self, message: SilaReal, toplevel_named_data_node: Optional[NamedDataNode] = None) -> float:
        return message.value

    def to_message(
        self, value: Union[float, int], toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> SilaReal:
        if isinstance(value, int):
            value = float(value)  # get meaningful errors for impossible conversions, e.g. on overflows
        return self.message_type(value=value)

    @staticmethod
    def from_string(value: str) -> float:
        # regex: https://www.w3.org/TR/xmlschema11-2/#nt-decimalRep
        if not re.match(r"^([+-])?([0-9]+(\.[0-9]*)?|\.[0-9]+)$", value):
            raise ValueError(f"Cannot parse as decimal: '{value}'")
        return float(value)



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_node import NamedNode
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class NamedDataNode(NamedNode):
    data_type: DataType

    def __init__(self, fdl_node, parent_feature: Feature, parent_namespace):
        super().__init__(fdl_node)
        self.data_type = DataType.from_fdl_node(
            xpath_sila(fdl_node, "sila:DataType")[0], parent_feature, parent_namespace
        )
